# encoding: utf-8
#
# DEC 2016 - Altered tests to work with differing definitions
#
# SNAGGED FROM https://github.com/mewwts/addict/blob/62e8481a2a5c8520259809fc306698489975c9f8/test_addict.py WITH THE FOLLOWING LICENCE
#
# The MIT License (MIT)
#
# Copyright (c) 2014 Mats Julian Olsen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import unicode_literals

from copy import deepcopy
from unittest import skip

import mo_json
from mo_future import Mapping
from mo_testing.fuzzytestcase import FuzzyTestCase

import mo_dots
from mo_dots import Data, set_default, from_data, to_data

TEST_VAL = [1, 2, 3]
TEST_DICT = {"a": {"b": {"c": TEST_VAL}}}


class AddictTests(FuzzyTestCase):
    def test_set_one_level_item(self):
        some_dict = {"a": TEST_VAL}
        prop = Data()
        prop["a"] = TEST_VAL
        self.assertEqual(prop, some_dict)

    def test_set_two_level_items(self):
        some_dict = {"a": {"b": TEST_VAL}}
        prop = Data()
        prop["a"]["b"] = TEST_VAL
        self.assertEqual(prop, some_dict)

    def test_set_three_level_items(self):
        prop = Data()
        prop["a"]["b"]["c"] = TEST_VAL
        self.assertEqual(prop, TEST_DICT)

    def test_set_one_level_property(self):
        prop = Data()
        prop.a = TEST_VAL
        self.assertEqual(prop, {"a": TEST_VAL})

    def test_set_two_level_properties(self):
        prop = Data()
        prop.a.b = TEST_VAL
        self.assertEqual(prop, {"a": {"b": TEST_VAL}})

    def test_set_three_level_properties(self):
        prop = Data()
        prop.a.b.c = TEST_VAL
        self.assertEqual(prop, TEST_DICT)

    def test_init_with_dict(self):
        self.assertEqual(TEST_DICT, to_data(TEST_DICT))

    def test_init_with_kws(self):
        prop = Data(a=2, b={"a": 2}, c=[{"a": 2}])
        self.assertEqual(prop, {"a": 2, "b": {"a": 2}, "c": [{"a": 2}]})

    def test_init_with_list(self):
        kv_pairs = [("0", 1), ("1", 2), ("2", 3)]
        prop = to_data(dict(kv_pairs))
        self.assertEqual(prop, {"0": 1, "1": 2, "2": 3})

    def test_init_raises(self):
        def init1():
            Data(5)

        def init2():
            Data("a")

        def init3():
            Data({})

        def init4():
            Data([])

        self.assertRaises(Exception, init1)
        self.assertRaises(Exception, init2)
        self.assertRaises(Exception, init3)
        self.assertRaises(Exception, init4)

    def test_init_with_empty_stuff(self):
        a = to_data({})
        b = to_data([])
        self.assertEqual(a, {})
        self.assertEqual(b, [])

    def test_init_with_list_of_dicts(self):
        a = to_data({"a": [{"b": 2}]})
        self.assertIsInstance(a.a[0], Data)
        self.assertEqual(a.a[0].b, 2)

    def test_getitem(self):
        prop = to_data(TEST_DICT)
        self.assertEqual(prop["a"]["b"]["c"], TEST_VAL)

    def test_empty_getitem(self):
        prop = Data()
        prop.a.b.c
        self.assertEqual(prop, {})

    def test_getattr(self):
        prop = to_data(TEST_DICT)
        self.assertEqual(prop.a.b.c, TEST_VAL)

    def test_isinstance(self):
        self.assertIsInstance(Data(), Mapping)

    def test_str(self):
        prop = to_data(TEST_DICT)
        self.assertEqual(str(prop), str(TEST_DICT))

    def test_json(self):
        some_dict = TEST_DICT
        some_json = mo_json.value2json(some_dict)
        prop = Data()
        prop.a.b.c = TEST_VAL
        prop_json = mo_json.value2json(prop)
        self.assertEqual(some_json, prop_json)

    def test_delitem(self):
        prop = to_data({"a": 2})
        del prop["a"]
        self.assertEqual(prop, {})

    def test_delitem_nested(self):
        prop = to_data(deepcopy(TEST_DICT))
        del prop["a"]["b"]["c"]
        self.assertEqual(prop, {"a": {"b": {}}})

    def test_delattr(self):
        prop = to_data({"a": 2})
        del prop.a
        self.assertEqual(prop, {})

    def test_delattr_nested(self):
        prop = to_data(deepcopy(TEST_DICT))
        del prop.a.b.c
        self.assertEqual(prop, {"a": {"b": {}}})

    def test_delitem_delattr(self):
        prop = to_data(deepcopy(TEST_DICT))
        del prop.a["b"]
        self.assertEqual(prop, {"a": {}})

    def test_set_prop_invalid(self):
        prop = Data()
        prop.keys = 2
        prop.items = 3

        self.assertEqual(prop, {"keys": 2, "items": 3})

    @skip("dir(Data()) returns just the data attributes, not the Data class attributes")
    def test_dir(self):
        """
        ALL CLASS ATTRIBUTES SHOULD SHOW UP IN THE INSTANCES
        """
        key = "a"
        prop = to_data({key: 1})
        dir_prop = dir(prop)

        dir_dict = dir(Data)
        for d in dir_dict:
            self.assertTrue(d in dir_prop, d)

        self.assertTrue("__methods__" not in dir_prop)
        self.assertTrue("__members__" not in dir_prop)

    def test_dir_with_members(self):
        prop = to_data({"__members__": 1})
        dir(prop)
        self.assertTrue("__members__" in prop.keys())

    def test_from_data(self):
        nested = {"a": [{"a": 0}, 2], "b": {}, "c": 2}
        prop = to_data(nested)
        regular = from_data(prop)
        self.assertEqual(regular, prop)
        self.assertEqual(regular, nested)
        self.assertNotIsInstance(regular, Data)

        def get_attr():
            regular.a = 2

        self.assertRaises(AttributeError, get_attr)

        def get_attr_deep():
            regular["a"][0].a = 1

        self.assertRaises(AttributeError, get_attr_deep)

    def test_to_dict_with_tuple(self):
        nested = {"a": ({"a": 0}, {2: 0})}
        prop = to_data(nested)
        regular = from_data(prop)
        self.assertEqual(regular, prop)
        self.assertEqual(regular, nested)
        self.assertIsInstance(regular["a"], tuple)
        self.assertNotIsInstance(regular["a"][0], Data)

    def test_update(self):
        old = Data()
        old.child.a = "a"
        old.child.b = "b"
        old.foo = "c"

        new = Data()
        new.child.b = "b2"
        new.child.c = "c"
        new.foo.bar = True

        old = set_default({}, new, old)

        reference = {"foo": {"bar": True}, "child": {"a": "a", "c": "c", "b": "b2"}}

        self.assertEqual(old, reference)

    def test_update1_with_lists(self):
        org = Data()
        org.a = [1, 2, {"a": "superman"}]
        someother = Data()
        someother.b = [{"b": 123}]
        org = someother | org

        correct = {"a": [1, 2, {"a": "superman"}], "b": [{"b": 123}]}

        org = someother | org
        self.assertEqual(org, correct)
        self.assertIsInstance(org.b[0], Mapping)

    def test_update2_with_lists(self):
        org = Data()
        org.a = [1, 2, {"a": "superman"}]
        someother = Data()
        someother.b = [{"b": 123}]
        org |= someother

        correct = {"a": [1, 2, {"a": "superman"}], "b": [{"b": 123}]}

        org |= someother
        self.assertEqual(org, correct)
        self.assertIsInstance(org.b[0], Mapping)

    def test_update1_with_kws(self):
        org = Data(one=1, two=2)
        someother = Data(one=3)
        someother = dict(one=1, two=2) | someother
        self.assertEqual(org, someother)

    def test_update2_with_kws(self):
        org = Data(one=3, two=2)
        someother = Data(one=3)
        someother |= dict(one=1, two=2)
        self.assertEqual(org, someother)

    def test_update1_with_args_and_kwargs(self):
        expected = {"a": 1, "b": 2}
        org = Data()
        org = {"a": 1} | ({"a": 3, "b": 2} | org)
        self.assertEqual(org, expected)

    def test_update2_with_args_and_kwargs(self):
        expected = {"a": 3, "b": 2}
        org = Data()
        org |= {"a": 3, "b": 2}
        org |= {"a": 1}
        self.assertEqual(org, expected)

    def test_hook_in_constructor(self):
        a_dict = to_data(TEST_DICT)
        self.assertIsInstance(a_dict["a"], Data)

    def test_copy(self):
        class MyMutableObject(object):
            def __init__(self):
                self.attribute = None

        foo = MyMutableObject()
        foo.attribute = True

        a = Data()
        a.child.immutable = 42
        a.child.mutable = foo

        b = a.copy()

        # immutable object should not change
        b.child.immutable = 21
        self.assertEqual(a.child.immutable, 21)

        # mutable object should change
        b.child.mutable.attribute = False
        self.assertEqual(a.child.mutable.attribute, b.child.mutable.attribute)

        # changing child of b should not affect a
        b.child = "new stuff"
        self.assertTrue(isinstance(a.child, Data))

    def test_deepcopy(self):
        class MyMutableObject(object):
            def __init__(self):
                self.attribute = None

        foo = MyMutableObject()
        foo.attribute = True

        a = Data()
        a.child.immutable = 42
        a.child.mutable = foo

        b = deepcopy(a)

        # immutable object should not change
        b.child.immutable = 21
        self.assertEqual(a.child.immutable, 42)

        # mutable object should not change
        b.child.mutable.attribute = False
        self.assertTrue(a.child.mutable.attribute)

        # changing child of b should not affect a
        b.child = "new stuff"
        self.assertTrue(isinstance(a.child, Data))

    def test_add_on_empty_dict(self):
        d = Data()
        d.x.y += 1

        self.assertEqual(d.x.y, 1)

    def test_add_on_non_empty_dict(self):
        d = Data()
        d.x.y = "defined"

        def run():
            d.x += 1

        self.assertRaises(Exception, run)

    def test_add_on_non_empty_value(self):
        d = Data()
        d.x.y = 1
        d.x.y += 1

        self.assertEqual(d.x.y, 2)

    def test_add_on_unsupported_type(self):
        d = Data()
        d.x.y = "str"

        def test():
            d.x.y += 1

        self.assertRaises(TypeError, test)

    def test_init_from_zip(self):
        keys = ["a"]
        values = [42]
        d = mo_dots.zip(keys, values)
        self.assertEqual(d.a, 42)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#



import re
import sys
from math import isnan

from mo_dots.datas import Data, data_types, is_data
from mo_dots.lists import (
    FlatList,
    is_list,
    is_sequence,
    is_container,
    is_many,
    list_types,
    container_types,
    finite_types,
    last,
)
from mo_dots.nones import Null, NullType
from mo_dots.objects import DataObject
from mo_dots.utils import CLASS, SLOT, get_logger, get_module
from mo_future import (
    binary_type,
    generator_types,
    text,
    OrderedDict,
    none_type,
    flatten,
    first,
)
from mo_imports import export

_module_type = type(sys.modules[__name__])
_builtin_zip = zip
_get = object.__getattribute__
_set = object.__setattr__
_new = object.__new__
_dict_zip = zip


def inverse(d):
    """
    reverse the k:v pairs
    """
    output = {}
    for k, v in from_data(d).items():
        output[v] = output.get(v, [])
        output[v].append(k)
    return output


def coalesce(*args):
    # pick the first not null value
    # http://en.wikipedia.org/wiki/Null_coalescing_operator
    for a in args:
        if a != None:
            return to_data(a)
    return Null


def zip(keys, values):
    """
    CONVERT LIST OF KEY/VALUE PAIRS TO Data
    PLEASE `import mo_dots`, AND CALL `mo_dots.zip()`
    """
    output = Data()
    for k, v in _dict_zip(keys, values):
        output[k] = v
    return output


def missing(value):
    raise NotImplementedError("use is_missing")


def is_missing(t) -> bool:
    # RETURN True IF EFFECTIVELY NOTHING
    class_ = t.__class__
    if class_ in null_types:
        return True
    elif class_ in data_types:
        return False
    elif class_ in finite_types and not t:
        return True
    elif class_ is text and not t:
        return True
    else:
        return t == None


def exists(value) -> bool:
    return not is_missing(value)


ESCAPE_DOTS1 = re.compile(r"(^\.|\.$)")  # DOTS AT START/END
ESCAPE_DOTS2 = re.compile(r"(?<!^)\.(?!$)")  # INTERNAL DOTS
ILLEGAL_DOTS = re.compile(r"[^.]\.(?:\.\.)+")  # ODD DOTS ARE NOT ALLOWED
SPLIT_DOTS = re.compile(r"(?<!\.)\.(?!\.)")  # SINGLE DOTS
UNESCAPE_DOTS = re.compile(r"\x08|(?:\.\.)")  # ENCODED DOTS


def literal_field(field):
    """
    RETURN SAME WITH DOTS (`.`) ESCAPED
    """
    try:
        return ESCAPE_DOTS2.sub("..", ESCAPE_DOTS1.sub("\b", field))
    except Exception as e:
        get_logger().error("bad literal", e)


def unliteral_field(field):
    """
    DUE TO PATHOLOGY IN MY CODE WE HAVE A path WITH ESCAPED DOTS BUT WE WANT OT USE IT ON A dict, NOT A Data
    a = dict()
    b = Data(a)
    a[unliteral_field(k)]==b[k] (for all k)

    :param field: THE STRING TO DE-literal IZE
    :return: SIMPLER STRING
    """
    return UNESCAPE_DOTS.sub(".", field)


def tail_field(field):
    """
    RETURN THE FIRST STEP IN PATH, ALONG WITH THE REMAINING TAILf
    IN (first, rest) PAIR
    """
    if field == "." or field == None or field == "":
        return ".", "."
    elif "." in field:
        path = split_field(field)
        if path[0].startswith("."):
            return path[0], join_field(path[1:])
        return literal_field(path[0]), join_field(path[1:])
    else:
        return field, "."


def split_field(field):
    """
    RETURN field AS ARRAY OF DOT-SEPARATED FIELDS
    """
    if ILLEGAL_DOTS.search(field):
        get_logger().error("Odd number of dots is not allowed")
    if field.startswith(".."):
        remainder = field.lstrip(".")
        back = len(field) - len(remainder) - 1
        return [".."] * back + [UNESCAPE_DOTS.sub(".", k) for k in SPLIT_DOTS.split(remainder) if k]
    else:
        return [UNESCAPE_DOTS.sub(".", k) for k in SPLIT_DOTS.split(field) if k]


def join_field(path):
    """
    RETURN field SEQUENCE AS STRING
    """
    if path.__class__ in generator_types:
        path = list(path)

    if not path:
        return "."

    prefix = ""
    while True:
        try:
            i = path.index("..")
            if i == 0:
                prefix += "."
                path = path[1:]
            else:
                path = path[: i - 1] + path[i + 1 :]
        except ValueError:
            return ("." if prefix else "") + prefix + ".".join(literal_field(f) for f in path)


def concat_field(*fields):
    return join_field(flatten(split_field(f) for f in fields))


def startswith_field(field, prefix):
    """
    RETURN True IF field PATH STRING STARTS WITH prefix PATH STRING
    """
    if prefix == None:
        return False
    if prefix.startswith("."):
        return True
        # f_back = len(field) - len(field.strip("."))
        # p_back = len(prefix) - len(prefix.strip("."))
        # if f_back > p_back:
        #     return False
        # else:
        #     return True

    if field.startswith(prefix):
        lp = len(prefix)
        if len(field) == len(prefix) or field[lp] in (".", "\b") and field[lp + 1] not in (".", "\b"):
            return True
    return False


def endswith_field(field, suffix):
    """
    RETURN True IF field PATH STRING ENDS WITH suffix PATH STRING
    """
    if suffix == None:
        return False
    if suffix == ".":
        return True

    if field.endswith(suffix):
        ls = len(suffix)
        if len(field) == ls or field[-ls - 1] in (".", "\b") and field[-ls - 2] not in (".", "\b"):
            return True
    return False


def relative_field(field, parent):
    """
    RETURN field PATH WITH RESPECT TO parent
    """
    if parent == ".":
        return field

    field_path = split_field(field)
    parent_path = split_field(parent)
    common = 0
    for f, p in _builtin_zip(field_path, parent_path):
        if f != p:
            break
        common += 1

    tail = join_field(field_path[common:])
    if len(parent_path) <= common:
        return join_field(field_path[common:])

    dots = "." * (len(parent_path) - common)
    if tail == ".":
        return "." + dots
    else:
        return "." + dots + tail


def hash_value(v):
    if is_many(v):
        return hash_value(first(v))
    elif _get(v, CLASS) in data_types:
        return hash_value(first(v.values()))
    else:
        return hash(v)


def fromkeys(keys, value=None):
    if value == None:
        return Data()
    return dict_to_data(dict.fromkeys(keys, value))


def set_default(d, *dicts):
    """
    RECURSIVE MERGE OF MULTIPLE dicts MOST IMPORTANT FIRST

    UPDATES d WITH THE MERGE RESULT, WHERE MERGE RESULT IS DEFINED AS:
    FOR EACH LEAF, RETURN THE FIRST NOT-NULL LEAF VALUE

    :param dicts: dicts IN PRIORITY ORDER, HIGHEST TO LOWEST
    :return: d
    """
    agg = d if d or _get(d, CLASS) in data_types else {}
    for p in dicts:
        _set_default(agg, p, seen={})
    return to_data(agg)


def _set_default(d, default, seen=None):
    """
    ANY VALUE NOT SET WILL BE SET BY THE default
    THIS IS RECURSIVE
    """
    if default is None:
        return

    for k, default_value in default.items():
        raw_value = from_data(default_value)  # TWO DIFFERENT Dicts CAN SHARE id() BECAUSE THEY ARE SHORT LIVED
        if is_data(d):
            existing_value = d.get(k)
        else:
            existing_value = _get_attr(d, [k])

        if existing_value == None:
            if default_value != None:
                if _get(default_value, CLASS) in data_types:
                    df = seen.get(id(raw_value))
                    if df is not None:
                        _set_attr(d, [k], df)
                    else:
                        copy_dict = {}
                        seen[id(raw_value)] = copy_dict
                        _set_attr(d, [k], copy_dict)
                        _set_default(copy_dict, default_value, seen)
                else:
                    # ASSUME PRIMITIVE (OR LIST, WHICH WE DO NOT COPY)
                    try:
                        _set_attr(d, [k], default_value)
                    except Exception as e:
                        if PATH_NOT_FOUND not in e:
                            get_logger().error("Can not set attribute {{name}}", name=k, cause=e)
        elif is_list(existing_value) or is_list(default_value):
            _set_attr(d, [k], None)
            _set_attr(d, [k], listwrap(existing_value) + listwrap(default_value))
        elif (hasattr(existing_value, "__setattr__") or _get(existing_value, CLASS) in data_types) and _get(
            default_value, CLASS
        ) in data_types:
            df = seen.get(id(raw_value))
            if df is not None:
                _set_attr(d, [k], df)
            else:
                seen[id(raw_value)] = existing_value
                _set_default(existing_value, default_value, seen)


def _getdefault(obj, key):
    """
    obj ANY OBJECT
    key IS EXPECTED TO BE LITERAL (NO ESCAPING)
    TRY BOTH ATTRIBUTE AND ITEM ACCESS, OR RETURN Null
    """
    try:
        return obj[key]
    except Exception as f:
        pass

    if is_sequence(obj):
        return [_getdefault(o, key) for o in obj]

    try:
        if obj.__class__ is not dict:
            return getattr(obj, key)
    except Exception as f:
        pass

    try:
        if float(key) == round(float(key), 0):
            return obj[int(key)]
    except Exception as f:
        pass

    # TODO: FIGURE OUT WHY THIS WAS EVER HERE (AND MAKE A TEST)
    # try:
    #     return eval("obj."+text(key))
    # except Exception as f:
    #     pass
    return NullType(obj, key)


PATH_NOT_FOUND = "Path not found"
AMBIGUOUS_PATH_FOUND = "Path is ambiguous"


def set_attr(obj, path, value):
    """
    SAME AS object.__setattr__(), BUT USES DOT-DELIMITED path
    RETURN OLD VALUE
    """
    try:
        return _set_attr(obj, split_field(path), value)
    except Exception as cause:
        Log = get_logger()
        if PATH_NOT_FOUND in cause:
            Log.warning(PATH_NOT_FOUND + ": {{path}}", path=path, cause=cause)
        else:
            Log.error("Problem setting value", cause=cause)


def get_attr(obj, path):
    """
    SAME AS object.__getattr__(), BUT USES DOT-DELIMITED path
    """
    try:
        return _get_attr(obj, split_field(path))
    except Exception as cause:
        Log = get_logger()
        if PATH_NOT_FOUND in cause:
            Log.error(PATH_NOT_FOUND + ": {{path}}", path=path, cause=cause)
        else:
            Log.error("Problem setting value", cause=cause)


def _get_attr(obj, path):
    if not path:
        return obj

    attr_name = path[0]

    if isinstance(obj, _module_type):
        if attr_name in obj.__dict__:
            return _get_attr(obj.__dict__[attr_name], path[1:])
        elif attr_name in dir(obj):
            return _get_attr(obj[attr_name], path[1:])

        # TRY FILESYSTEM
        File = get_module("mo_files").File
        possible_error = None
        python_file = (File(obj.__file__).parent / attr_name).set_extension("py")
        python_module = File(obj.__file__).parent / attr_name / "__init__.py"
        if python_file.exists or python_module.exists:
            try:
                # THIS CASE IS WHEN THE __init__.py DOES NOT IMPORT THE SUBDIR FILE
                # WE CAN STILL PUT THE PATH TO THE FILE IN THE from CLAUSE
                if len(path) == 1:
                    # GET MODULE OBJECT
                    output = __import__(
                        obj.__name__ + str(".") + str(attr_name), globals(), locals(), [str(attr_name)], 0,
                    )
                    return output
                else:
                    # GET VARIABLE IN MODULE
                    output = __import__(
                        obj.__name__ + str(".") + str(attr_name), globals(), locals(), [str(path[1])], 0,
                    )
                    return _get_attr(output, path[1:])
            except Exception as e:
                Except = get_module("mo_logs.exceptions.Except")
                possible_error = Except.wrap(e)

        # TRY A CASE-INSENSITIVE MATCH
        matched_attr_name = lower_match(attr_name, dir(obj))
        if not matched_attr_name:
            get_logger().warning(
                PATH_NOT_FOUND + "({{name|quote}}) Returning None.", name=attr_name, cause=possible_error,
            )
        elif len(matched_attr_name) > 1:
            get_logger().error(AMBIGUOUS_PATH_FOUND + " {{paths}}", paths=attr_name)
        else:
            return _get_attr(obj, matched_attr_name + path[1:])

    try:
        obj = obj[int(attr_name)]
        return _get_attr(obj, path[1:])
    except Exception:
        pass

    try:
        obj = getattr(obj, attr_name)
        return _get_attr(obj, path[1:])
    except Exception:
        pass

    try:
        obj = obj[attr_name]
        return _get_attr(obj, path[1:])
    except Exception as f:
        return NullType(obj, attr_name)


def _set_attr(obj_, path, value):
    obj = _get_attr(obj_, path[:-1])
    if obj is None:
        # DELIBERATE USE OF `is`: WE DO NOT WHAT TO CATCH Null HERE (THEY CAN BE SET)
        get_logger().error(PATH_NOT_FOUND + " tried to get attribute of None")

    attr_name = path[-1]

    # ACTUAL SETTING OF VALUE
    try:
        old_value = _get_attr(obj, [attr_name])
        old_type = _get(old_value, CLASS)
        if old_value == None or old_type in (bool, int, float, text, binary_type):
            old_value = None
            new_value = value
        elif value == None:
            new_value = None
        else:
            new_value = _get(old_value, CLASS)(value)  # TRY TO MAKE INSTANCE OF SAME CLASS
    except Exception:
        old_value = None
        new_value = value

    try:
        setattr(obj, attr_name, new_value)
        return old_value
    except Exception as e:
        try:
            obj[attr_name] = new_value
            return old_value
        except Exception as f:
            get_logger().error(PATH_NOT_FOUND, cause=[f, e])


def lower_match(value, candidates):
    return [v for v in candidates if v.lower() == value.lower()]


def dict_to_data(d):
    """
    FASTEST WAY TO MAKE Data, DO NOT CHECK TYPE
    :param d: dict
    :return: Data
    """
    m = _new(Data)
    _set(m, SLOT, d)
    return m


def list_to_data(v):
    """
    to_data, BUT WITHOUT CHECKS
    """
    output = _new(FlatList)
    _set(output, SLOT, v)
    return output


def to_data(v=None) -> object:
    """
    WRAP AS Data OBJECT FOR DATA PROCESSING: https://github.com/klahnakoski/mo-dots/tree/dev/docs
    :param v:  THE VALUE TO WRAP
    :return:  Data INSTANCE
    """

    type_ = _get(v, CLASS)

    if type_ in (dict, OrderedDict):
        m = _new(Data)
        _set(m, SLOT, v)
        return m
    elif type_ is none_type:
        return Null
    elif type_ is tuple:
        return list_to_data(v)
    elif type_ is list:
        return list_to_data(v)
    elif type_ in generator_types:
        return list_to_data(list(from_data(vv) for vv in v))
    else:
        return v


wrap = to_data


def leaves_to_data(value):
    """
    dict WITH DOTS IN KEYS IS INTERPRETED AS A PATH
    """
    return to_data(_leaves_to_data(value))


wrap_leaves = leaves_to_data


def _leaves_to_data(value):
    """
    RETURN UNWRAPPED STRUCTURES
    """
    if value == None:
        return None

    class_ = _get(value, CLASS)
    if class_ in (text, binary_type, int, float):
        return value

    if class_ in data_types:
        if class_ is Data:
            value = from_data(value)

        output = {}
        for key, value in value.items():
            value = _leaves_to_data(value)

            if key == "":
                get_logger().error("key is empty string.  Probably a bad idea")

            seq = split_field(key)
            if not seq:
                if not output:
                    output = value
                continue
            if not is_data(output):
                output = {}
            d = output
            for k in seq[:-1]:
                e = d.get(k, None)
                if not is_data(e):
                    e = d[k] = {}
                d = e

            if value == None:
                d.pop(seq[-1], None)
            else:
                d[seq[-1]] = value
        return output

    if hasattr(value, "__iter__"):
        return [_leaves_to_data(v) for v in value]

    return value


def from_data(v):
    if v is None:
        return None
    _type = _get(v, CLASS)
    if _type is NullType:
        return None
    elif _type is Data:
        d = _get(v, SLOT)
        return d
    elif _type is FlatList:
        return _get(v, SLOT)
    elif _type is DataObject:
        return _get(v, SLOT)
    elif _type in generator_types:
        return (from_data(vv) for vv in v)
    elif _type is float:
        if isnan(v):
            return None
        return v
    return v


unwrap = from_data


def listwrap(value):
    """
    PERFORMS THE FOLLOWING TRANSLATION
    None -> []
    value -> [value]
    [...] -> [...]  (unchanged list)

    ## MOTIVATION ##
    OFTEN IT IS NICE TO ALLOW FUNCTION PARAMETERS TO BE ASSIGNED A VALUE,
    OR A list-OF-VALUES, OR NULL.  CHECKING FOR WHICH THE CALLER USED IS
    TEDIOUS.  INSTEAD WE CAST FROM THOSE THREE CASES TO THE SINGLE CASE
    OF A LIST

    # BEFORE
    def do_it(a):
        if a is None:
            return
        if not isinstance(a, list):
            a=[a]
        for x in a:
            # do something

    # AFTER
    def do_it(a):
        for x in listwrap(a):
            # do something

    """
    if value == None:
        return FlatList()
    elif is_list(value):
        if isinstance(value, list):
            return list_to_data(value)
        else:
            return value
    elif is_many(value):
        return list_to_data(list(value))
    else:
        return list_to_data([from_data(value)])


def unwraplist(v):
    """
    LISTS WITH ZERO AND ONE element MAP TO None AND element RESPECTIVELY
    """
    if is_list(v):
        if len(v) == 0:
            return None
        elif len(v) == 1:
            return from_data(v[0])
        else:
            return from_data(v)
    else:
        return from_data(v)


def tuplewrap(value):
    """
    INTENDED TO TURN lists INTO tuples FOR USE AS KEYS
    """
    if value == None:
        return tuple()
    elif is_many(value):
        return tuple(tuplewrap(v) if is_sequence(v) else v for v in value)
    else:
        return (from_data(value),)


def is_null(t):
    # RETURN True IF EFFECTIVELY NOTHING
    class_ = t.__class__
    if class_ in null_types:
        return True
    else:
        try:
            return t == None
        except Exception:
            return False


def is_not_null(t):
    # RETURN True IF EFFECTIVELY SOMETHING
    class_ = t.__class__
    if class_ in null_types:
        return False
    elif class_ in data_types:
        return True
    elif class_ in finite_types and t:
        return True
    else:
        return t != None


null_types = (none_type, NullType)

# EXPORT
export("mo_dots.nones", to_data)
export("mo_dots.nones", null_types)
export("mo_dots.nones", get_attr)

export("mo_dots.datas", list_to_data)
export("mo_dots.datas", dict_to_data)
export("mo_dots.datas", to_data)
export("mo_dots.datas", from_data)
export("mo_dots.datas", coalesce)
export("mo_dots.datas", _getdefault)
export("mo_dots.datas", hash_value)
export("mo_dots.datas", listwrap)
export("mo_dots.datas", literal_field)
export("mo_dots.datas", null_types)
export("mo_dots.datas", concat_field)

export("mo_dots.lists", list_to_data)
export("mo_dots.lists", to_data)
export("mo_dots.lists", coalesce)
export("mo_dots.lists", from_data)
export("mo_dots.lists", hash_value)

export("mo_dots.objects", list_to_data)
export("mo_dots.objects", dict_to_data)
export("mo_dots.objects", to_data)
export("mo_dots.objects", from_data)
export("mo_dots.objects", get_attr)
export("mo_dots.objects", set_attr)
export("mo_dots.objects", set_default)



# encoding: utf-8

# SNAGED FROM https://github.com/pawelzny/dotty_dict/blob/98984911a61ae9f1aa4da3f6c4808da991b89847/tests/test_dotty_value_access.py
# UNDER THe MIT LICENSE
from unittest import skip

from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_dots import to_data


class TestDottyValueAccess(FuzzyTestCase):
    def setUp(self):
        self.dot = to_data({
            "flat_key": "flat value",
            "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here",},},},
        })

    def test_access_flat_value(self):
        self.assertEqual(self.dot["flat_key"], "flat value")

    def test_raise_key_error_if_key_does_not_exist(self):
        result = self.dot["not_existing"]
        self.assertTrue(result == None)

    def test_access_deep_nested_value(self):
        self.assertEqual(self.dot["deep.nested"], 12)

    def test_access_middle_nested_value(self):
        self.assertEqual(self.dot["deep.deeper.ridiculous"], {"hell": "is here"})

    def test_set_flat_value(self):
        self.dot["new_flat"] = "super flat"
        self.assertIn("new_flat", self.dot)

    def test_set_deep_nested_value(self):
        self.dot["deep.new_key"] = "new value"
        self.assertIn("new_key", self.dot["deep"])

    def test_set_new_deeply_nested_value(self):
        self.dot["other.chain.of.keys"] = True
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here",},},},
                "other": {"chain": {"of": {"keys": True,},},},
            },
        )

    def test_dotty_has_flat_key(self):
        self.assertIn("flat_key", self.dot)

    def test_dotty_has_deeply_nested_key(self):
        self.assertIn("deep.nested", self.dot)

    def test_dotty_has_not_flat_key(self):
        self.assertNotIn("some_key", self.dot)

    def test_dotty_has_not_deeply_nested_key(self):
        self.assertNotIn("deep.other.chain", self.dot)

    def test_has_in(self):
        result = "deep.deeper.secret" in self.dot
        self.assertTrue(result)

    def test_has_not_in(self):
        result = "deep.other" in self.dot
        self.assertFalse(result)

    def test_delete_flat_key(self):
        del self.dot["flat_key"]
        self.assertEqual(
            self.dot, {"deep": {"nested": 12, "deeper": {"secret": "abcd", "ridiculous": {"hell": "is here",},},},},
        )

    def test_delete_nested_key(self):
        del self.dot["deep.deeper.secret"]
        self.assertEqual(
            self.dot,
            {"flat_key": "flat value", "deep": {"nested": 12, "deeper": {"ridiculous": {"hell": "is here",},},},},
        )

    def test_raise_key_error_on_delete_not_existing_key(self):
        del self.dot["deep.deeper.key"]

    def test_set_value_with_escaped_separator(self):
        self.dot[r"deep.deeper.escaped..dot_key"] = "it works!"
        self.assertEqual(
            self.dot,
            {
                "flat_key": "flat value",
                "deep": {
                    "nested": 12,
                    "deeper": {"escaped.dot_key": "it works!", "secret": "abcd", "ridiculous": {"hell": "is here",},},
                },
            },
        )

    def test_get_value_with_escaped_separator(self):
        dot = to_data({
            "flat_key": "flat value",
            "deep": {
                "nested": 12,
                "deeper": {"escaped.dot_key": "it works!", "secret": "abcd", "ridiculous": {"hell": "is here",},},
            },
        })
        result = dot[r"deep.deeper.escaped..dot_key"]
        self.assertEqual(result, "it works!")

    def test_get_value_with_escaped_escape_separator1(self):
        dot = to_data({
            "flat_key": "flat value",
            "deep": {
                "nested": 12,
                "deeper": {
                    "escaped\\": {"dot_key": "it works!",},
                    "secret": "abcd",
                    "ridiculous": {"hell": "is here",},
                },
            },
        })
        result = dot["deep.deeper.escaped\\.dot_key"]
        self.assertEqual(result, "it works!")

    def test_get_value_with_escaped_escape_separator2(self):
        dot = to_data({"deep": {"deeper": {"escaped.": {"dot_key": "it works!"}},},})
        result = dot["deep.deeper.escaped\b.dot_key"]
        self.assertEqual(result, "it works!")

    def test_string_digit_key(self):
        dot = to_data({"field": {"1": "one", "5": "five"}})

        dict_one = dot["field.1"]
        dict_five = dot["field.5"]

        self.assertEqual(dict_one, "one")
        self.assertEqual(dict_five, "five")

    def test_integer_keys(self):
        dot = to_data({"field": {1: "one", 5: "five"}})

        dict_one = dot["field.1"]
        dict_five = dot["field.5"]

        self.assertEqual(dict_one, "one")
        self.assertEqual(dict_five, "five")

    def test_data_gathering_with_int(self):
        dot = to_data({
            "2": "string_value",
            2: "int_value",
            "nested": {"2": "nested_string_value", 3: "nested_int_value"},
        })

        dict_string = dot["2"]
        dict_int = dot[2]  # KEYS MUST BE STRINGS. THEY WILL BE CAST
        nested_dict_string = dot["nested.2"]
        nested_dict_int = dot["nested.3"]

        self.assertEqual(dict_string, "string_value")
        self.assertEqual(dict_int, "string_value")
        self.assertEqual(nested_dict_string, "nested_string_value")
        self.assertEqual(nested_dict_int, "nested_int_value")

    @skip("mo-dots only deals with string keys")
    def test_non_standard_key_types(self):
        dot = Data({3.3: "float", True: "bool", None: "None", "nested": {4.4: "nested_float"}}, separator=",",)

        dict_float = dot[3.3]
        dict_bool = dot[True]
        dict_none = dot[None]
        nested_dict_float = dot["nested,4.4"]
        self.assertEqual(dict_float, "float")
        self.assertEqual(dict_bool, "bool")
        self.assertEqual(dict_none, "None")
        self.assertEqual(nested_dict_float, "nested_float")



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#



from collections import OrderedDict
from copy import deepcopy
from datetime import date, datetime
from decimal import Decimal

from mo_future import (
    binary_type,
    generator_types,
    get_function_arguments,
    get_function_defaults,
    text,
    Mapping,
)
from mo_imports import export, expect

from mo_dots.datas import register_data, Data, _iadd
from mo_dots.lists import FlatList
from mo_dots.nones import NullType, Null
from mo_dots.utils import CLASS, SLOT

get_attr, set_attr, list_to_data, dict_to_data, to_data, from_data, set_default = expect(
    "get_attr", "set_attr", "list_to_data", "dict_to_data", "to_data", "from_data", "set_default"
)

_new = object.__new__
_get = object.__getattribute__
_set = object.__setattr__
WRAPPED_CLASSES = set()


class DataObject(Mapping):
    """
    TREAT AN OBJECT LIKE DATA
    """

    __slots__ = [SLOT]

    def __init__(self, obj):
        _set(self, SLOT, obj)

    def __getattr__(self, item):
        obj = _get(self, SLOT)
        output = get_attr(obj, item)
        return object_to_data(output)

    def __setattr__(self, key, value):
        obj = _get(self, SLOT)
        set_attr(obj, key, value)

    def __getitem__(self, item):
        obj = _get(self, SLOT)
        output = get_attr(obj, item)
        return object_to_data(output)

    def __or__(self, other):
        return set_default({}, self, other)

    def __ror__(self, other):
        return to_data(other) | self

    def __add__(self, other):
        return to_data(_iadd(_iadd({}, self), other))

    def __radd__(self, other):
        return to_data(_iadd(_iadd({}, other), self))

    def get(self, item):
        obj = _get(self, SLOT)
        output = get_attr(obj, item)
        return object_to_data(output)

    def keys(self):
        obj = _get(self, SLOT)
        try:
            return obj.__dict__.keys()
        except Exception as e:
            raise e

    def items(self):
        obj = _get(self, SLOT)
        try:
            for k, v in obj.__dict__.items():
                yield k, object_to_data(v)
        except Exception as e:
            for k in dir(obj):
                if k.startswith("__"):
                    continue
                yield k, object_to_data(getattr(obj, k, None))

    def __deepcopy__(self, memodict={}):
        output = {}
        for k, v in self.items():
            output[k] = from_data(deepcopy(v))
        return dict_to_data(output)

    def __data__(self):
        return self

    def __iter__(self):
        return (k for k in self.keys())

    def __str__(self):
        obj = _get(self, SLOT)
        return str(obj)

    def __len__(self):
        obj = _get(self, SLOT)
        return len(obj)

    def __call__(self, *args, **kwargs):
        obj = _get(self, SLOT)
        return obj(*args, **kwargs)


register_data(DataObject)


def object_to_data(v):
    try:
        if v == None:
            return Null
    except Exception:
        pass

    type_ = _get(v, CLASS)

    if type_ is (dict, OrderedDict):
        m = _new(Data)
        _set(m, SLOT, v)
        return m
    elif type_ is tuple:
        return list_to_data(v)
    elif type_ is list:
        return list_to_data(v)
    elif type_ in (Data, DataObject, FlatList, NullType):
        return v
    elif type_ in (text, binary_type, int, float, Decimal, datetime, date):
        return v
    elif type_ in generator_types:
        return (to_data(vv) for vv in v)

    return DataObject(v)


datawrap = object_to_data


class DataClass(object):
    """
    ALLOW INSTANCES OF class_ TO ACT LIKE dicts
    ALLOW CONSTRUCTOR TO ACCEPT @override
    """

    def __init__(self, class_):
        WRAPPED_CLASSES.add(class_)
        self.class_ = class_
        self.constructor = class_.__init__

    def __call__(self, *args, **kwargs):
        settings = to_data(kwargs).settings

        params = get_function_arguments(self.constructor)[1:]
        func_defaults = get_function_defaults(self.constructor)
        if not func_defaults:
            defaults = {}
        else:
            defaults = {k: v for k, v in zip(reversed(params), reversed(func_defaults))}

        ordered_params = dict(zip(params, args))

        output = self.class_(**params_pack(params, ordered_params, kwargs, settings, defaults))
        return DataObject(output)


def params_pack(params, *args):
    settings = {}
    for a in args:
        for k, v in a.items():
            k = text(k)
            if k in settings:
                continue
            settings[k] = v

    output = {str(k): from_data(settings[k]) for k in params if k in settings}
    return output


export("mo_dots.lists", "object_to_data", object_to_data)
export("mo_dots.lists", "datawrap", object_to_data)



# encoding: utf-8

# SNAGED FROM https://github.com/pawelzny/dotty_dict/blob/98984911a61ae9f1aa4da3f6c4808da991b89847/tests/test_dotty_basics.py
# UNDER THe MIT LICENSE
from __future__ import unicode_literals

from unittest import skip

from mo_dots import to_data, Data
from mo_testing.fuzzytestcase import FuzzyTestCase


class TestDottyBasics(FuzzyTestCase):
    def test_create_empty_instance(self):
        dot = to_data({})
        self.assertEqual(dot, {})

    def test_create_non_empty_instance(self):
        plain_dict = {"not": "empty"}

        dot = to_data(plain_dict)
        self.assertEqual(dot, plain_dict)
        self.assertIsNot(dot, plain_dict)

    def test_raise_attr_error_if_input_is_not_dict(self):
        with self.assertRaises(Exception):
            Data(["not", "valid"])

    def test_two_dotty_with_the_same_input_should_be_equal(self):
        first = to_data({"is": "valid"})
        second = to_data({"is": "valid"})

        self.assertEqual(first, second)

    def test_two_dotty_with_different_input_should_not_be_equal(self):
        first = to_data({"counter": 1})
        second = to_data({"counter": 2})

        self.assertNotEqual(first, second)

    def test_plain_dict_and_dotty_wrapper_should_be_equal(self):
        plain = {"a": 1, "b": 2}
        dot = to_data(plain)
        self.assertEqual(dot, plain)

    def test_dotty_and_not_mapping_instance_should_not_be_equal(self):
        dot = to_data({"a": 1, "b": 2})
        self.assertNotEqual(dot, [("a", 1), ("b", 2)])
        self.assertNotEqual(dot, ("a", 1))
        self.assertNotEqual(dot, {1, 2, 3})
        self.assertNotEqual(dot, 123)
        self.assertNotEqual(dot, "a:1, b:2")

    def test_pop_with_default_value(self):
        dot = to_data({})
        self.assertEqual(dot.pop("does.not.exist", None), None)
        self.assertEqual(dot.pop("does.not.exist", 55), 55)
        self.assertEqual(dot.pop("does.not.exist", "my_value"), "my_value")
